#!/usr/bin/perl

=head1 delete-system.pl

Deletes a virtual system or systems

Virtual systems can also be deleted from the command line. This should be done
with extreme care, as you will not get any confirmation before deletion
happens, and all data on the system will be lost forever. The system to
destroy is specified with the C<--host> parameter.

If the C<--keep> flag is given, collected information about the system such
as bandwidth and memory usage will be preserved by Cloudmin. If a new system
is later created with the same ID, the collected information will be associated
with it.

If the C<--keep-disks> flag is given, any virtual disks or filesystem associated
with the system will be preserved on its host until manually removed or
connected to another system.

=head1 Example usage

  cloudmin delete-system --host xennoupdate.home
  Deleting Xen system xennoupdate.home ..
  .. deleted successfully

=cut

package server_manager;
$main::no_acl_check++;
$ENV{'WEBMIN_CONFIG'} ||= "/etc/webmin";
$ENV{'WEBMIN_VAR'} ||= "/var/webmin";
if ($0 =~ /^(.*)\/[^\/]+$/) {
	chdir($pwd = $1);
	}
else {
	chop($pwd = `pwd`);
	}
$0 = "$pwd/delete-system.pl";
require './server-manager-lib.pl';
$< == 0 || die "$0 must be run as root";
&set_all_text_print();
@OLDARGV = @ARGV;
&usage_if_cloudmin_replica();

# Parse command line
while(@ARGV > 0) {
	my $a = shift(@ARGV);
	if ($a eq "--host") {
		push(@hosts, shift(@ARGV));
		}
	elsif ($a eq "--owner") {
		push(@owners, shift(@ARGV));
		}
	elsif ($a eq "--keep") {
		$keep = 1;
		}
	elsif ($a eq "--keep-disks") {
		$keepdisks = 1;
		}
	elsif ($a eq "--multiline") {
		$multiline = 1;
		}
	elsif ($a eq "--override-lock") {
		$override = 1;
		}
	else {
		&usage("Unknown parameter $a");
		}
	}

# Work out who to list
@servers = &find_managed_servers(0, \@hosts, [ ], [ ], [ ], \@owners);
@servers || &usage("No systems specified");
&can_action_usage(\@servers, "delete"); 

# Check if they can all be deleted
foreach $s (@servers) {
	if ($s->{'manager'} eq 'real' || $s->{'manager'} eq 'this') {
		push(@cannot, $s);
		&usage("$s->{'host'} is not a virtual system that ".
		       "can be deleted");
		}
	}
if (@cannot) {
	&usage("The following systems are not virtual and cannot be deleted : ".
	       join(" ", map { $_->{'host'} } @cannot));
	}

# Check for dependencies
@deps = &find_server_dependencies(@servers);
if (@deps) {
	&usage("The following systems have dependencies and cannot be ".
	       "deleted : ".join(" ", map { $_->{'host'} } @deps));
	}

# Check if leaving disks is allowed
if ($keepdisks) {
	&can_action(\@servers, "devices") ||
		&usage("You are not allowed to use the --keep-disks flag");
	}

# Do it!
foreach $s (@servers) {
	if ($override) {
		# Kill any existing locks
		my $lock = &test_lock_server($s);
		if ($lock) {
			print "Killing process $lock with lock on $s->{'host'} ..\n";
			&kill_logged('KILL', $lock);
			print ".. done\n";
			}
		}

	&obtain_lock_server($s);
	$tfunc = "type_".$s->{'manager'}."_desc";
	$tdesc = &$tfunc();

	# Run pre-deletion script
	&run_system_script($s, 'before', 'delete') || next;

	&delete_all_domains($s);

	print "Deleting $tdesc system $s->{'host'} ..\n";
	$dfunc = "type_".$s->{'manager'}."_delete";
	$err = &$dfunc($s, $keepdisks);
	if ($err) {
		print ".. deletion failed : $err\n\n";
		}
	else {
		&clear_ssh_host_key($s);
		&delete_managed_server($s, $keep);
		print ".. deleted successfully\n\n";

		# Run post-deletion script
		&run_system_script($s, 'after', 'delete');
		}
	&release_lock_server($s);
	}
&cloudmin_api_log(\@OLDARGV);

sub usage
{
print "$_[0]\n\n" if ($_[0]);
print "Deletes one or more managed systems.\n";
print "\n";
print "cloudmin delete-system [--host name]*\n";
print "                       [--owner name]*\n";
print "                       [--keep]\n";
print "                       [--keep-disks]\n";
print "                       [--override-lock]\n";
exit(1);
}

